package main

import (
	"io"
	"log"
	"net"
	"sync"
)

const (
	listenAddr = "127.0.0.1:7000"
	serverAddr = "127.0.0.1:8000"
)

func main() {
	err := run()
	if err != nil {
		log.Fatal(err)
	}
}

func run() error {
	addr, err := net.ResolveTCPAddr("tcp", listenAddr)
	if err != nil {
		return err
	}
	ln, err := net.ListenTCP("tcp", addr)
	if err != nil {
		return err
	}
	defer ln.Close()

	return acceptConnections(ln)
}

func acceptConnections(ln *net.TCPListener) error {
	// Accept connections on local TCP listener.
	for {
		local, err := ln.AcceptTCP()
		if err != nil {
			return err
		}
		go func() {
			defer local.Close()
			log.Printf("begin local %s", local.RemoteAddr())
			err := handleConnection(local)
			log.Printf("end   local %s %v", local.RemoteAddr(), err)
		}()
	}
}

func handleConnection(local *net.TCPConn) error {
	// Connect to remote server.
	server, err := net.Dial("tcp", serverAddr)
	if err != nil {
		return err
	}
	defer server.Close()

	var wg sync.WaitGroup
	var outErr, inErr error
	wg.Add(2)
	go func() {
		defer wg.Done()
		_, outErr = io.Copy(server, local)
		local.CloseRead()
		server.Close()
	}()
	go func() {
		defer wg.Done()
		_, inErr = io.Copy(local, server)
		local.CloseWrite()
	}()
	wg.Wait()

	err = outErr
	if err == nil {
		err = inErr
	}
	return err
}
