package main

import (
	"io"
	"log"
	"net"
	"sync"
)

const (
	listenAddr  = "127.0.0.1:8000"
	forwardAddr = "127.0.0.1:7001"
)

func main() {
	err := run()
	if err != nil {
		log.Fatal(err)
	}
}

func run() error {
	ln, err := net.Listen("tcp", listenAddr)
	if err != nil {
		return err
	}
	defer ln.Close()

	return acceptConnections(ln)
}

func acceptConnections(ln net.Listener) error {
	// Accept connections from remote client.
	for {
		client, err := ln.Accept()
		if err != nil {
			return err
		}
		go func() {
			defer client.Close()
			log.Printf("begin client %s", client.RemoteAddr())
			err := handleConnection(client)
			log.Printf("end   client %s %v", client.RemoteAddr(), err)
		}()
	}
}

func handleConnection(client net.Conn) error {
	// Connect to forward TCP port.
	addr, err := net.ResolveTCPAddr("tcp", forwardAddr)
	if err != nil {
		return err
	}
	forward, err := net.DialTCP("tcp", nil, addr)
	if err != nil {
		return err
	}
	defer forward.Close()

	var wg sync.WaitGroup
	var outErr, inErr error
	wg.Add(2)
	go func() {
		defer wg.Done()
		_, inErr = io.Copy(client, forward)
		forward.CloseRead()
		client.Close()
	}()
	go func() {
		defer wg.Done()
		_, outErr = io.Copy(forward, client)
		forward.CloseWrite()
	}()
	wg.Wait()

	err = outErr
	if err == nil {
		err = inErr
	}
	return err
}
