package main

import (
	"encoding/binary"
	"io"
	"log"
	"net"
	"sync"
	"time"

	"github.com/xtaci/kcp-go/v5"
	"github.com/xtaci/smux"

	"www.bamsoftware.com/git/turbotunnel-paper.git/example/turbotunnel/turbotunnel"
)

const (
	listenAddr = "127.0.0.1:7000"
	serverAddr = "127.0.0.1:8000"
)

func main() {
	err := run()
	if err != nil {
		log.Fatal(err)
	}
}

type emptyAddr struct{}

func (addr emptyAddr) Network() string { return "empty" }
func (addr emptyAddr) String() string  { return "empty" }

type tcpDialPacketConn struct {
	*turbotunnel.QueuePacketConn
}

func newTCPDialPacketConn(addr string) (*tcpDialPacketConn, error) {
	conn, err := net.Dial("tcp", addr)
	if err != nil {
		return nil, err
	}
	qpconn := turbotunnel.NewQueuePacketConn(nil, 1*time.Minute)
	sessionID := turbotunnel.NewSessionID()
	log.Printf("session ID %s", sessionID)
	_, err = conn.Write(sessionID[:])
	if err != nil {
		return nil, err
	}
	// Send loop.
	go func() {
		for p := range qpconn.OutgoingQueue(emptyAddr{}) {
			err := binary.Write(conn, binary.BigEndian, uint16(len(p)))
			if err != nil {
				return
			}
			_, err = conn.Write(p)
			if err != nil {
				return
			}
		}
	}()
	// Receive loop.
	go func() {
		for {
			var n uint16
			err := binary.Read(conn, binary.BigEndian, &n)
			if err != nil {
				return
			}
			p := make([]byte, n)
			_, err = io.ReadFull(conn, p)
			if err != nil {
				return
			}
			qpconn.QueueIncoming(p, sessionID)
		}
	}()
	return &tcpDialPacketConn{qpconn}, nil
}

func run() error {
	addr, err := net.ResolveTCPAddr("tcp", listenAddr)
	if err != nil {
		return err
	}
	ln, err := net.ListenTCP("tcp", addr)
	if err != nil {
		return err
	}
	defer ln.Close()

	pconn, err := newTCPDialPacketConn(serverAddr)
	if err != nil {
		return err
	}
	defer pconn.Close()

	conn, err := kcp.NewConn2(emptyAddr{}, nil, 0, 0, pconn)
	if err != nil {
		return err
	}
	defer conn.Close()

	smuxConfig := smux.DefaultConfig()
	smuxConfig.Version = 2
	smuxConfig.KeepAliveTimeout = 1 * time.Minute
	sess, err := smux.Client(conn, smuxConfig)
	if err != nil {
		return err
	}
	defer sess.Close()

	return acceptConnections(ln, sess)
}

func acceptConnections(ln *net.TCPListener, sess *smux.Session) error {
	// Accept connections on local TCP listener.
	for {
		local, err := ln.AcceptTCP()
		if err != nil {
			return err
		}
		go func() {
			defer local.Close()
			log.Printf("begin local %s", local.RemoteAddr())
			err := handleConnection(local, sess)
			log.Printf("end   local %s %v", local.RemoteAddr(), err)
		}()
	}
}

func handleConnection(local *net.TCPConn, sess *smux.Session) error {
	stream, err := sess.OpenStream()
	if err != nil {
		return err
	}
	defer stream.Close()

	var wg sync.WaitGroup
	var outErr, inErr error
	wg.Add(2)
	go func() {
		defer wg.Done()
		_, outErr = io.Copy(stream, local)
		local.CloseRead()
		stream.Close()
	}()
	go func() {
		defer wg.Done()
		_, inErr = io.Copy(local, stream)
		local.CloseWrite()
	}()
	wg.Wait()

	err = outErr
	if err == nil {
		err = inErr
	}
	return err
}
