package main

import (
	"encoding/binary"
	"io"
	"log"
	"net"
	"sync"
	"time"

	"github.com/xtaci/kcp-go/v5"
	"github.com/xtaci/smux"

	"www.bamsoftware.com/git/turbotunnel-paper.git/example/turbotunnel/turbotunnel"
)

const (
	listenAddr  = "127.0.0.1:8000"
	forwardAddr = "127.0.0.1:7001"
)

type tcpListenPacketConn struct {
	*turbotunnel.QueuePacketConn
}

func newTCPListenPacketConn(addr string) (*tcpListenPacketConn, error) {
	ln, err := net.Listen("tcp", addr)
	if err != nil {
		return nil, err
	}
	qpconn := turbotunnel.NewQueuePacketConn(ln.Addr(), 1 * time.Minute)
	conn := &tcpListenPacketConn{qpconn}
	go conn.acceptConnections(ln)
	return conn, nil
}

func (c *tcpListenPacketConn) acceptConnections(ln net.Listener) error {
	for {
		conn, err := ln.Accept()
		if err != nil {
			return err
		}
		go func() {
			defer conn.Close()
			log.Printf("begin TCP conn %s", conn.RemoteAddr())
			err := c.handleConnection(conn)
			log.Printf("end   TCP conn %s %v", conn.RemoteAddr(), err)
		}()
	}
}

func (c *tcpListenPacketConn) handleConnection(conn net.Conn) error {
	var sessionID turbotunnel.SessionID
	_, err := io.ReadFull(conn, sessionID[:])
	if err != nil {
		return err
	}

	var wg sync.WaitGroup
	var outErr, inErr error
	wg.Add(2)
	go func() {
		defer wg.Done()
		for p := range c.QueuePacketConn.OutgoingQueue(sessionID) {
			err := binary.Write(conn, binary.BigEndian, uint16(len(p)))
			if err != nil {
				return
			}
			_, err = conn.Write(p)
			if err != nil {
				return
			}
		}
	}()
	go func() {
		defer wg.Done()
		for {
			var n uint16
			err := binary.Read(conn, binary.BigEndian, &n)
			if err != nil {
				return
			}
			p := make([]byte, n)
			_, err = io.ReadFull(conn, p)
			if err != nil {
				return
			}
			c.QueuePacketConn.QueueIncoming(p, sessionID)
		}
	}()
	wg.Wait()

	err = outErr
	if err == nil {
		err = inErr
	}
	return err
}

func main() {
	err := run()
	if err != nil {
		log.Fatal(err)
	}
}

func run() error {
	pconn, err := newTCPListenPacketConn(listenAddr)
	if err != nil {
		return err
	}

	ln, err := kcp.ServeConn(nil, 0, 0, pconn)
	if err != nil {
		return err
	}
	defer ln.Close()

	return acceptSessions(ln)
}

func acceptSessions(ln *kcp.Listener) error {
	// Accept connections from remote client.
	for {
		client, err := ln.AcceptKCP()
		if err != nil {
			return err
		}
		go func() {
			defer client.Close()
			log.Printf("begin client %s", client.RemoteAddr())
			err := acceptStreams(client)
			log.Printf("end   client %s %v", client.RemoteAddr(), err)
		}()
	}
}

func acceptStreams(conn *kcp.UDPSession) error {
	smuxConfig := smux.DefaultConfig()
	smuxConfig.Version = 2
	smuxConfig.KeepAliveTimeout = 1 * time.Minute
	sess, err := smux.Server(conn, smuxConfig)
	if err != nil {
		return err
	}
	defer sess.Close()

	for {
		stream, err := sess.AcceptStream()
		if err != nil {
			return err
		}
		go func() {
			defer stream.Close()
			log.Printf("begin stream %v", stream.ID())
			err := handleStream(stream)
			log.Printf("end   stream %v %v", stream.ID(), err)
		}()
	}
}

func handleStream(stream *smux.Stream) error {
	// Connect to forward TCP port.
	addr, err := net.ResolveTCPAddr("tcp", forwardAddr)
	if err != nil {
		return err
	}
	forward, err := net.DialTCP("tcp", nil, addr)
	if err != nil {
		return err
	}
	defer forward.Close()

	var wg sync.WaitGroup
	var outErr, inErr error
	wg.Add(2)
	go func() {
		defer wg.Done()
		_, inErr = io.Copy(stream, forward)
		forward.CloseRead()
		stream.Close()
	}()
	go func() {
		defer wg.Done()
		_, outErr = io.Copy(forward, stream)
		forward.CloseWrite()
	}()
	wg.Wait()

	err = outErr
	if err == nil {
		err = inErr
	}
	return err
}
